# sqlite3vfs.tcl -- TclSqlite3 Virtual File System driver
# Copyright (C) 2005 Tiago Dionizio. All Rights Reserved.
# Tiago Dionizio <tngd81@gmail.com>
#
# 18apr05 tngd  1       First working version
# 16apr05 tngd  1       Initial version
# CAVEATS:
#
#   Concurrency:
#       * Only one writer can exist at the same time
#       * Many readers can access the database
#       * One writer can coexist with many writers, although he
#       will not be able to commit changes until all readers end
#       the transaction.
#       * When there is a writer active and another writer wants
#       to start writing too, he will get a locked database. To
#       prevent this situation and use the busy timer, try to use
#       the RunLock with an IMMEDIATE lock.
# TODO:
#   * Reconfigure any of the options given to Mount
#   * Filesystem integrity check (find unreachable objects)
# Visible interface:
#
#   vfs::sqlite3::Mount file local ?args ...? : fd
#       Mount the database in 'file' as directory 'local'.
#       Returns the file descriptor for the new filesystem.
#
#       args:
#           -timeout value      # lock timeout
#           -busy script        # busy lock handler (overrides timeout)
#           -key key            # encryption key
#           -compress level     # compression level (0 = store)
#
#   vfs::sqlite3::Unmount fd local
#       Unmounts the filesystem identified by 'fd' mounted in
#       'local' path.
#
#   vfs::sqlite3::Runlock fd script ?locktype?
#       Make changes in the filesystem in a transaction which
#       may improve performance when doing multiple operations.
#       The 'locktype' can be DEFERRED or IMMEDIATE (values that
#       make sense to use here) see the sqlite3 documentation
#       for more information on transaction types.
#       The script 'script' will be run and the transaction
#       will be closed when it finishes execution. If it raises
#       an error, the transaction will be rolled back and
#       invalidating any changes made during execution.
package require vfs 1.3             ;# virtual filesystem
package require sqlite3 3           ;# the database engine
catch {package require zlib 1.1}    ;# for compression
package provide sqlite3vfs 1
package provide vfs::sqlite3 1
namespace eval vfs::sqlite3 {
    # place all internal procs and variables in here
    namespace eval _ {
        # database names are numbers
        variable seq 0
        # posix error to raise when an error occours
        variable perror
        # transaction lock count
        variable lock
        # filesystem options
        variable option
        array set lock {}
        array set perror {}
        array set option {}
        proc clearerror {db} {
            variable perror
            set perror($db) {}
        }
        proc throw {db code} {
            variable perror
            set perror($db) $code
            ::vfs::filesystem posixerror $::vfs::posix($code)
        }
        proc rethrow {db} {
            variable perror
            set e $perror($db)
            if {$e ne ""} {
                ::vfs::filesystem posixerror $::vfs::posix($e)
            }
        }
    }
}
# Mount the sqlite3 database <file> as directory <local>
# args = ?-key key?
proc vfs::sqlite3::Mount {dbfile local args} {
    #::vfs::log "Mount $dbfile $local $args"
    if {$dbfile eq "" || $dbfile eq ":memory:"} {
        # temporary filesystem
        set dbfile ":memory:"
    } else {
        # follow links
        set dbfile [::vfs::filesystem fullynormalize $dbfile]
    }
    set db [eval [list _::mount $dbfile] $args]
    ::vfs::filesystem mount $local [list [namespace current]::handler $db]
    ::vfs::RegisterMount $local [list [namespace current]::Unmount $db]
    return $db
}
# Unmount the sqlite3 database file
proc vfs::sqlite3::Unmount {db local} {
    ::vfs::filesystem unmount $local
    _::unmount $db
}
## vfshandler subcmd root relative actualpath args...
 #     The first argument specifies the operation to perform on behalf of
 #     the filesystem code in the tcl core, the remainder specify the file
 #     path on which to operate, in different forms, and parts, and any
 #     additional arguments which may be required to carry out the action.
 #
 #     To demonstrate the treatment of a path by the generic layer we use
 #     "C:/foo/bar/mount.zip/xxx/yyy" as an example and additionally assume
 #     that the following conditions are true:
 #
 #        1. "mount.zip" is a zip archive which has been mounted on top of
 #           itself,
 #        2. said zip archive contains a file with path "xxx/yyy",
 #        3. the current working directory of the application is inside of
 #           directory "xxx",
 #        4. and the command executed is file exists yyy.
 #
 #     The file separator between root and relative is omitted. Most
 #     filesystem operations need only the relative argument for their
 #     correct operation, but some actually require the other parts of the
 #     path.
 #
 #     subcmd
 #         This argument of the handler can be one of the following access,
 #         createdirectory, deletefile, fileattributes, matchindirectory,
 #         open, removedirectory, stat, or utime.
 #
 #         The generic layer expects that the subcommands of a handler
 #         signal error conditions by calling vfs::filesystem posixerror
 #         with the appropriate posix error code instead of throwing a tcl
 #         error. If the latter is done nevertheless it will be treated as
 #         an unknown posix error.
 #
 #         There are three exceptions to the rule above: If any of open
 #         (when an interpreter is given), matchindirectory, and
 #         fileattributes (for a set or get operation only) throw a tcl
 #         error, this error will be passed up to the caller of the
 #         filesystem command which invoked the handler. Note that this
 #         does not preclude the ability of these subcommands to use the
 #         command vfs::filesystem posixerror to report more regular
 #         filesystem errors.
 #
 #     root
 #         Part of the specification of the path to operate upon. It
 #         contains the part of the path which lies outside this
 #         filesystem's mount point. For example outlined above its
 #         value will be "C:/foo/bar/mount.zip".
 #
 #     relative
 #         Part of the specification of the path to operate upon. It
 #         contains the part of the path which lies inside this
 #         filesystem. For example outlined above its value will be
 #         "xxx/yyy".
 #
 #     actualpath
 #         Part of the specification of the path to operate upon. It
 #         contains the original (unnormalized) name of the path which
 #         was used in the current command wherever it originated (in
 #         Tcl or C). For example outlined above its value will be "yyy".
 ##
# start a transaction or increase lock count
# type can be DEFERRED, IMMEDIATE, EXCLUSIVE
# type will only affect first lock. locks can not be promoted
proc vfs::sqlite3::_::begin {db {type DEFERRED}} {
    variable lock
    upvar 0 lock($db) l
    if {$l == 0} {
        $db eval "BEGIN $type"
    }
    incr l
    return $l
}
# close a transaction.
proc vfs::sqlite3::_::end {db {rollback 0}} {
    variable lock
    upvar 0 lock($db) l
    if {$l <= 0} {
        error "Filesystem already unlocked."
    }
    incr l -1
    if {$l == 0} {
        if {$rollback} {
            $db eval {ROLLBACK}
        } else {
            $db eval {END}
        }
    }
    return $l
}
# run a script inside a transaction.
# not allowing exclusive access because we want give as much
# time to other readers to access the filesystem.
proc vfs::sqlite3::RunLock {db script {locktype DEFERRED}} {
    if {$locktype ne "IMMEDIATE" && $locktype ne "DEFERRED"} {
        error "locktype must be DEFERRED or IMMEDIATE"
    }
    if {[catch {_::begin $db $locktype}]} {
        error "Failed to lock filesystem"
    }
    set code [catch {uplevel 1 $script} res]
    if {$code} {
        set ecode $::errorCode
        set einfo $::errorInfo
        catch {_::end $db 1}
        # errno information seems to be lost if used
        return -code $code -errorinfo $einfo -errorcode $ecode $res
    } else {
        if {[catch {_::end $db}]} {
            error "Failed to release lock on filesystem"
        }
        return $res
    }
}
proc vfs::sqlite3::handler {db subcmd root relative actualpath args} {
    #::vfs::log "*** handler :: $subcmd $db $root -$relative- $actualpath $args"
    # Choose appropriate locking mechanism.
    # Hopefully, this will help with concurrency issues; when aquiring
    # reserved locks when there is already a reader, will cause the
    # busy handler to be invoqued and wait for a while to try to get
    # write lock on the database. Can be looked at as a potencial
    # blocking operation. This is intended to avoid as many locking
    # problems as possible, without failing to execute later sql statements.
    # Note that these functions are not recursive, so there should be
    # no problems here, unless user calls RunLock proc with an inadequate
    # lock (such as shared lock when a reserved one is needed and there is
    # another writer doing some work).
    switch -exact -- $subcmd {
        createdirectory -
        removedirectory -
        deletefile -
        open -
        utime   { set lock IMMEDIATE ;# planning to write }
        default { set lock DEFERRED ;# reading only }
    }
    _::begin $db $lock
    _::clearerror $db
    set code [catch {eval [list _::$subcmd $db $root $relative $actualpath] $args} res]
    if {$code} {
        if {$::vfs::debug > 0} {
            set ecode {}
            set einfo {}
            if {[info exists ::errorInfo]} {
                set einfo $::errorInfo
            }
            if {[info exists ::errorCode]} {
                set ecode $::errorCode
            }
            ::vfs::log "handler: error :: $code"
            ::vfs::log $einfo
            ::vfs::log $ecode
        }
        catch {_::end $db 1}
        # rethrow posix error if there is one
        _::rethrow $db
        # or else, throw an EIO error
        _::throw $db EIO
    } else {
        if {[catch {_::end $db} res2]} {
            _::throw $db EIO
        }
        return $res
    }
}
## vfshandler access root relative actualpath mode
 #     Signal a posix error if the specified access mode (an integer
 #     number) is not compatible with the file or directory described
 #     by the path. The generic layer will ignore any non-empty return
 #     value.
 #
 #     The command vfs::accessMode (see section HANDLER ENVIRONMENT)
 #     can be used to convert the integer mode into an easier to check
 #     string value.
 ##
proc vfs::sqlite3::_::access {db root relative actualpath mode args} {
    set attr [find $db $relative]
    if {[llength $attr] == 0} {
        if {[::vfs::accessMode $mode] ne "W"} {
            throw $db ENOENT
        }
    }
}
## vfshandler createdirectory root relative actualpath
 #     Create a directory with the given name. The command can
 #     assume that all sub-directories in the path exist and are
 #     valid, and that the actual desired path does not yet exist
 #     (Tcl takes care of all of that for us).
 ##
proc vfs::sqlite3::_::createdirectory {db root relative actualpath args} {
    # concurrency checking
    if {[llength [find $db $relative]] > 0} {
        # already exists. do nothing.
        return
    }
    set dir [file tail $relative]
    # find parent id
    set p_dir [file dirname $relative]
    set p_id [lindex [find $db $p_dir] 0]
    # create directory
    #::vfs::log "creating directory $p_id,$dir"
    set ctime [clock seconds]
    if {[catch {$db eval {INSERT INTO dirs(parent,name,ctime) VALUES($p_id,$dir,$ctime)}}]} {
        throw $db EIO
    }
}
## vfshandler deletefile root relative actualpath
 #     Delete the given file.
 ##
proc vfs::sqlite3::_::deletefile {db root relative actualpath args} {
    set id [lindex [find $db $relative] 0]
    if {$id ne ""} {
        $db eval {DELETE FROM files WHERE id = $id}
    }
}
## vfshandler fileattributes root relative actualpath ?index? ?value?
 #     The command has to return a list containing the names of all
 #     acceptable attributes, if neither index nor value were specified.
 #
 #     The command has to return the value of the index'th attribute if
 #     the index is specified, but not the value. The attributes are
 #     counted in the same order as their names appear in the list
 #     returned by a call where neither index nor value were specified.
 #     The first attribute is has the index 0.
 #
 #     The command has to set the value of the index'th attribute to
 #     value if both index and value were specified for the call.
 ##
proc vfs::sqlite3::_::fileattributes {db root relative actualpath {index {}} {value {}} args} {
    switch -- [llength $args] {
        0 {
            # list strings
            return [::vfs::listAttributes]
        }
        1 {
            # get value
            set index [lindex $args 0]
            return [::vfs::attributesGet $root $relative $index]
        }
        2 {
            # set value
            #if {$::mode($db) == "readonly"} {
            #    vfs::filesystem posixerror $::vfs::posix(EROFS)
            #}
            #set index [lindex $args 0]
            #set val [lindex $args 1]
            return [::vfs::attributesSet $root $relative $index $val]
        }
    }
}
## vfshandler matchindirectory root relative actualpath pattern types
 #     Return the list of files or directories in the given path which
 #     match the glob pattern and are compatible with the specified
 #     list of types. The specified path is always the name of an
 #     existing directory.
 #
 #     Note: As Tcl generates requests for directory-only matches from
 #     the filesystems involved when performing any type of recursive
 #     globbing this subcommand absolutely has to handle such (and
 #     file-only) requests correctly or bad things (TM) will happen.
 #
 #     The commands vfs::matchDirectories and vfs::matchFiles (see
 #     section HANDLER ENVIRONMENT) can aid the implementation greatly
 #     in this task.
 ##
proc vfs::sqlite3::_::getdirs {db relative pattern} {
    set id [lindex [find $db $relative] 0]
    if {$id eq ""} {
        return
    }
    set dirs {}
    $db eval {SELECT name FROM dirs WHERE parent = $id} {
        if {[string match $pattern $name]} {
            lappend dirs $name
        }
    }
    return $dirs
}
proc vfs::sqlite3::_::getfiles {db relative pattern} {
    set id [lindex [find $db $relative] 0]
    if {$id eq ""} {
        return
    }
    set files {}
    $db eval {SELECT name FROM files WHERE parent = $id} {
        if {[string match $pattern $name]} {
            lappend files $name
        }
    }
    return $files
}
proc vfs::sqlite3::_::matchindirectory {db root relative actualpath pattern types args} {
    #::vfs::log "matchindirectory $db $root $relative $actualpath $pattern $types $args"
    set match {}
    if {$pattern eq ""} {
        # match single file
        if {[llength [find $db $relative]] > 0} {
            set actual_match [list $actualpath]
        } else {
            set actual_match {}
        }
    } else {
        if {[::vfs::matchDirectories $types]} {
            set match [concat $match [getdirs $db $relative $pattern]]
        }
        if {[::vfs::matchFiles $types]} {
            set match [concat $match [getfiles $db $relative $pattern]]
        }
        # checking for read and write types will not make a different at the moment.
        # all files can be written to and read from.
        set actual_match {}
        foreach p [lsort $match] {
            lappend actual_match [file join $actualpath $p]
        }
    }
    return $actual_match
}
## vfshandler open root relative actualpath mode permissions
 #     Either returns a list describing the successfully opened file,
 #     or throws an error describing how the operation failed.
 #
 #     The list returned upon success contains at least one and at
 #     most two elements. The first, obligatory, element is always
 #     the handle of the channel which was created to allow access
 #     to the contents of the file.
 #
 #     If specified the second element will be interpreted as a
 #     callback, i.e. a command prefix. This prefix will always be
 #     executed as is, i.e. without additional arguments. Any
 #     required arguments have to be returned as part of the result
 #     of the call to open.
 #
 #     If present the specified callback will be evaluated just before
 #     the channel is closed by the generic filesystem layer. The
 #     callback itself must not call close.
 #
 #     The channel however is live enough to allow seek and read
 #     operations. In addition all available data will have been
 #     flushed into it already. This means, for example, that the
 #     callback can seek to the beginning of the said channel, read
 #     its contents and then store the gathered data elsewhere. In
 #     other words, this callback is not only crucial to the cleanup
 #     of any resources associated with an opened file, but also for
 #     the ability to implement a filesystem which can be written to.
 #
 #     Under normal circumstances return code and any errors thrown
 #     by the callback itself are ignored. In that case errors have
 #     to be signaled asychronously, for example by calling bgerror.
 #     However if, through a call of the subcommand internalerror,
 #     an error handling script has been specified for the file
 #     system, all errors thrown here will be passed to that script
 #     for further action.
 #
 #     mode
 #         can be any of r, w, a, w+, or a+.
 #
 #     permissions
 #         determines the native mode the openend file is created
 #         with. Relevant only of the open mode actually requests
 #         the creation of a non-existing file, i.e. is not r.
 ##
proc vfs::sqlite3::_::writefile {db relative fd} {
    # lock database?
    # if file does not exist anymore contents are discarded
    set attr [find $db $relative]
    set id [lindex $attr 0]
    set type [lindex $attr 1]
    if {$type eq "file"} {
        variable option
        set compress $option($db,compress)
        fconfigure $fd -translation binary
        seek $fd 0
        set contents [read $fd]
        set size [string bytelength $contents]
        set mtime [clock seconds]
        if {$compress} {
            # compress contents
            catch {
                set ncontents [zlib compress $contents $compress]
                # replace only if size is smaller
                if {[string bytelength $ncontents] < $size} {
                    set contents $ncontents
                }
            }
        }
        set sql {UPDATE files SET size = $size, contents = $contents, mtime = $mtime WHERE id = $id}
        begin $db IMMEDIATE
        if {[catch {$db eval $sql}]} {
            catch {end $db 1}
            throw $db EIO
        } else {
            end $db
        }
    }
}
proc vfs::sqlite3::_::open {db root relative actualpath mode permissions args} {
    # return a list of two elements:
    # 1. first element is the Tcl channel name which has been opened
    # 2. second element (optional) is a command to evaluate when
    #  the channel is closed.
    set attr [find $db $relative]
    set id [lindex $attr 0]
    set type [lindex $attr 1]
    set size 0
    set contents {}
    if {$type eq "directory"} {
        throw $db EISDIR
    }
    # read initial contents if they exist
    if {$id eq ""} {
        # create empty file as needed
        switch -glob -- $mode {
            {} -
            r* {
                throw $db ENOENT
            }
            a* -
            w* {
                # create empty file
                set ctime [clock seconds]
                set dirname [file dirname $relative]
                set name [file tail $relative]
                set p_id [lindex [find $db $dirname] 0]
                # i assume parent directory exists always
                $db eval {
                    INSERT INTO files(parent, name, size, ctime, mtime, contents)
                    VALUES($p_id,$name,0,$ctime,$ctime,'')
                }
                # if it failt, operation will throw an error (as it should)
                # and raise EIO error
            }
        }
    } else {
        # read initial contents or truncate file
        switch -glob -- $mode {
            {} -
            r* -
            a* {
                $db eval {SELECT size, contents FROM files WHERE id = $id} {break}
                # uncompress if needed
                if {$size != [string bytelength $contents]} {
                    catch {set contents [zlib decompress $contents]}
                }
            }
            w* {
                $db eval {UPDATE files SET size = 0, contents = '' WHERE id = $id}
            }
        }
    }
    # open channel and place initial contents inside.
    set fd [::vfs::memchan]
    fconfigure $fd -translation binary
    puts -nonewline $fd $contents
    seek $fd 0
    fconfigure $fd -translation auto
    #
    switch -glob -- $mode {
        r+ -
        a* -
        w* {
            if {[string match "a*" $mode]} {
                seek $fd 0 end
            }
            return [list $fd [list [namespace current]::writefile $db $relative $fd]]
        }
        default {
            return $fd
        }
    }
}
## vfshandler removedirectory root relative actualpath recursive
 #     Delete the given directory. Argument recursive is a boolean.
 #     If the specified value is true then even if the directory is
 #     non-empty, an attempt has to be made to recursively delete it
 #     and its contents. If the spcified value is false and the
 #     directory is non-empty, a posix error (EEXIST) has to be thrown.
 ##
proc vfs::sqlite3::_::removedirectory_r {db base_id} {
    # find child directories and delete them
    foreach cid [$db eval {SELECT id FROM dirs WHERE parent = $base_id}] {
        removedirectory_r $db $cid
    }
    # now delete the files
    $db eval {DELETE FROM files WHERE parent = $base_id}
    # and finally, delete the base directory
    $db eval {DELETE FROM dirs WHERE id = $base_id}
}
proc vfs::sqlite3::_::removedirectory {db root relative actualpath recursive args} {
    # get dir name
    set dir [file tail $relative]
    set id [lindex [find $db $relative] 0]
    if {$recursive} {
        # delete everything
        if {[catch {removedirectory_r $db $id}]} {
            throw $db EIO
        }
        # all done
    } else {
        # see if directory is not empty
        set child [$db eval {
            SELECT id FROM dirs WHERE parent = $id UNION
            SELECT id FROM files WHERE parent = $id LIMIT 1
        }]
        #::vfs::log "children found: $child"
        if {$child ne ""} {
            throw $db EEXIST
        } else {
            # remove the directory
            if {[catch {$db eval {DELETE FROM dirs WHERE id = $id}}]} {
                throw $db EIO
            }
        }
    }
}
## vfshandler stat root relative actualpath
 #     The result has to be a list of keys and values, in a format
 #     acceptable to the builtin command array set. It describes the
 #     contents of a stat structure. The order of the keys in the
 #     list is not important.
 #
 #     Given this the subcommand should use something like
 #
 #         return [list dev 0 type file mtime 1234 ...].
 #
 #     as the last command of its implementation.
 #
 #     The following keys and their values have to be supplied by the
 #     filesystem:
 #
 #     dev
 #         A long integer number, the device number of the path stat
 #         was called for.
 #
 #     ino
 #         A long integer number, the inode number of the path stat
 #         was called for. Each path handled by the filesystem should
 #         be uniquely identified by the combination of device and
 #         inode number. Violating this principle will cause
 #         higher-level algorithms which(have to) keep track of device
 #         and inode information to fail in all manners possible.
 #
 #         An example of such an algorithm would be a directory walker
 #         using device/inode information to keep itself out of
 #         infinite loops generated through symbolic links. Returning
 #         non-unique device/inode information will most likely cause
 #         such a walker to skip over paths under the wrong assumption
 #         of having them seen already.
 #
 #     mode
 #         An integer number, the access mode of the path. It is this
 #         mode which is checked by the subcommand access.
 #
 #     nlink
 #         A long integer number, the number of hard links to the
 #         specified path.
 #
 #     uid
 #         A long integer number, the id of the user owning the
 #         virtual path.
 #
 #     gid
 #         A long integer number, the id of the user group the
 #         virtual path belongs to.
 #
 #     size
 #         A long integer number, the true size of the virtual path,
 #         in bytes.
 #
 #     atime
 #         A long integer number, the time of the latest access to
 #         the path, in seconds since the epoch. Convertible into a
 #         readable date/time by the command clock format.
 #
 #     mtime
 #         A long integer number, the time of the latest modification
 #         of the path, in seconds since the epoch. Convertible into a
 #         readable date/time by the command clock format.
 #
 #     ctime
 #         A long integer number, the time of the path was created, in
 #         seconds since the epoch. Convertible into a readable
 #         date/time by the command clock format.
 #
 #     type
 #         A string, either directory, or file, describing the type of
 #         the given path.
 ##
proc vfs::sqlite3::_::stat {db root relative actualpath args} {
    set fid [find $db $relative]
    # not found
    if {[llength $fid] == 0} {
        throw $db ENOENT
    }
    set type [lindex $fid 1]
    set id [lindex $fid 0]
    array set info {}
    if {$type eq "directory"} {
        $db eval {SELECT count(*) + 1 AS nlink FROM files WHERE parent = $id} {break}
        $db eval {SELECT ctime FROM dirs WHERE id = $id} {break}
        set info(dev)   0
        set info(ino)   0
        set info(mode)  0777
        set info(nlink) $nlink
        set info(uid)   0
        set info(gid)   0
        set info(size)  0
        set info(atime) $ctime
        set info(mtime) $ctime
        set info(ctime) $ctime
        set info(type)  $type
    } else {
        # get the rest of the information of the file
        $db eval {SELECT size, ctime, mtime FROM files WHERE id = $id} {break}
        set info(dev)   0
        set info(ino)   0
        set info(mode)  0777
        set info(nlink) 1
        set info(uid)   0
        set info(gid)   0
        set info(size)  $size
        set info(atime) $mtime
        set info(mtime) $mtime
        set info(ctime) $ctime
        set info(type)  $type
    }
    return [array get info]
}
## vfshandler utime root relative actualpath actime mtime
 #     Set the access and modification times of the given file (these
 #     are read with stat).
 ##
proc vfs::sqlite3::_::utime {db root relative actualpath actime mtime args} {
    set id [lindex [find $db $relative]]
    if {$id ne ""} {
        $db eval {UPDATE files SET mtime = $mtime WHERE id = $id}
    }
}
## HANDLER ENVIRONMENT
 #
 # The implementation of a filesystem handler can rely on the
 # existence of the following utility commands:
 #
 # vfs::accessMode mode
 #     This commands converts an access mode given as integer
 #     into a string, one of F, X, W, XW, R, RX, and RW.
 #
 # vfs::matchDirectories types
 #     Checks if the glob types specification ask for the inclusion
 #     of directories. Returns a boolean result. true is returned
 #     if types does ask for directories, else false.
 #
 # vfs::matchFiles types
 #     Checks if the glob types specification ask for the inclusion
 #     of files. Returns a boolean result. true is returned if types
 #     does ask for directories, else false.
 #
 # vfs::matchCorrectTypes types filelist ?inDir?
 #     Returns that subset of the filelist which are compatible with
 #     the types given. The elements of filelist are either absolute
 #     paths, or names of files in the directory indir. The latter
 #     interpretation is taken if and only if the argument indir is
 #     specified.
 ##
# Mount the database
proc vfs::sqlite3::_::mount {file args} {
    variable seq
    variable lock
    variable option
    set key {}          ;# encryption key
    set timeout 5000    ;# timeout for writing operations
    set compress 0      ;# compression level (0 = store)
    set busy {}         ;# busy handler (overrides timeout)
    # parse arguments
    foreach {arg value} $args {
        switch -exact -- $arg {
            -key        { set key $value        }
            -timeout    { set timeout $value    }
            -compress   { set compress $value   }
            -busy       { set busy $value       }
        }
    }
    # argument validation
    # key not validated
    # busy handler not validated
    # timeout must be positive
    if {$timeout < 0 || $timeout eq ""} {
        error "invalid timeout value"
    }
    # compression level
    if {$compress > 9 || $timeout < 0} {
        error "compression level must be between 0 and 9"
    }
    # new sequence
    set db sqlite3fs[incr seq]
    #open the database file
    if {$key ne ""} {
        sqlite3 $db $file -key $key
    } else {
        sqlite3 $db $file
    }
    # key not saved... of course
    set option($db,compress) $compress
    set option($db,busy) $busy
    set option($db,timeout) $timeout
    # initialize the database as needed
    if {[catch {init $db} err]} {
        set errorInfo $::errorInfo
        # close database
        unset option($db,compress)
        unset option($db,busy)
        unset option($db,timeout)
        $db close
        error "Failed to open database $file." $errorInfo
    }
    # lock count
    set lock($db) 0
    return $db
}
# Unmount the database
proc vfs::sqlite3::_::unmount {db} {
    variable lock
    variable perror
    variable option
    unset lock($db)
    unset perror($db)
    unset option($db,compress)
    unset option($db,busy)
    unset option($db,timeout)
    $db close
}
# database busy handler - for debug
proc vfs::sqlite3::_::busyhandler {db tries} {
    #::vfs::log "busy: $db :: $tries"
    # retry for a short amount of time (configurable in the future)
    if {$tries > 50} {
        # abort
        return 1
    }
    # sleep 100 ms
    after 100
    # try again
    return 0
}
# trace sql execution for debug
proc vfs::sqlite3::_::tracehandler {db sql} {
    #::vfs::log "trace: $db :: $sql"
}
# get the column names of a table
proc vfs::sqlite3::_::column_names {db table} {
    # match dirs columns
    set cols {}
    $db eval "PRAGMA table_info($table)" {
        lappend cols $name
    }
    return $cols
}
# Initialize the database filesystem
proc vfs::sqlite3::_::init {db} {
    # set busy handler so we don't get imediate errors
    # while trying to access the database
    #$db busy [list [namespace current]::busyhandler $db]
    #$db trace [list [namespace current]::tracehandler $db]
    #$db eval {PRAGMA synchronous = OFF}
    variable option
    if {$option($db,busy) ne ""} {
        $db busy $option($db,busy)
    } else {
        $db timeout $option($db,timeout)
    }
    # see if needed tables exist
    set tables [$db eval {
        SELECT
            name
        FROM
            sqlite_master
        WHERE
            type = 'table' AND
            (name = 'dirs' OR name = 'files')
        ORDER BY name
    }]
    if {$tables eq {dirs files}} {
        # the tables exist. see if the columns are correct
        # match dirs columns
        set cols [column_names $db dirs]
        if {$cols ne {id parent name ctime}} {
            error "Column names for dirs table do not match -> $cols"
        }
        # match files columns
        set cols [column_names $db files]
        if {$cols ne {id parent name size ctime mtime contents}} {
            error "Column names for dirs table do not match -> $cols"
        }
    } else {
        # create initial tables
        $db eval {PRAGMA auto_vacuum = 1}
        $db eval {BEGIN EXCLUSIVE}
        $db eval {CREATE TABLE dirs(
            id INTEGER PRIMARY KEY,
            parent INTEGER,
            name TEXT,
            ctime INTEGER,
            UNIQUE(parent, name)
        )}
        $db eval {CREATE TABLE files(
            id INTEGER PRIMARY KEY,
            parent INTEGER,
            name TEXT,
            size INTEGER,
            ctime INTEGER,
            mtime INTEGER,
            contents BLOB,
            UNIQUE(parent, name)
        )}
        set ctime [clock seconds]
        $db eval {INSERT INTO dirs(id,name,ctime) VALUES(0,'',$ctime)}
        $db eval {COMMIT}
    }
    # all is ok
}
# Find the id of a given path.
# If found, returns the id of the path and the type of the path (dir or file).
proc vfs::sqlite3::_::find {db relative} {
    set sp [file split $relative]
    set tail [lindex $sp end]
    set parent 0
    set type directory
    foreach dir [lrange $sp 0 end-1] {
        set id [$db eval {SELECT id FROM dirs WHERE parent = $parent AND name = $dir LIMIT 1}]
        if {$id == ""} {
            return
        }
        set parent $id
        #::vfs::log "find::loop $parent - $dir"
    }
    if {$tail eq "." || $tail eq ":" || $tail eq ""} {
        #::vfs::log "$parent directory"
        return [list $parent directory]
    } else {
        # child directory?
        #::vfs::log "set id <$db onecolumn <SELECT id FROM dirs WHERE parent = $parent and name = $tail>>"
        set id [$db eval {SELECT id FROM dirs WHERE parent = $parent and name = $tail LIMIT 1}]
        if {$id ne ""} {
            return [list $id directory]
        }
        # File
        set id [$db eval {SELECT id FROM files WHERE parent = $parent AND name = $tail LIMIT 1}]
        if {$id ne ""} {
            return [list $id file]
        } else {
            return
        }
    }
}
